{{
*************************************
* boot module
*************************************

{{

How this code is designed to work:

1. Bootloader uses serial port, eeprom, and SDIO card by default.
It lets users run a spin program from SDIO card.

2. Configuration of pins for SDIO card in EEPROM configuration space starting at $C0.
This way the EEPROM configuration space configuration can be shared with applications. 

3. The first time the code boots or any time the EEPROM is reprogrammed,
the user will be prompted for the SDIO base pin. Then the bootconf.bin
program on SD card will be booted if it exists.

4. Normal boot will be interrupted if a character is received after "Bootloader Started"
and before loading program. This will cause the boot filename to be "bootconf.bin" if
the character is "C", otherwise user will be prompted for program from SD card directory.

}}


OBJ
  pd    : "PinDefs"
  bd    : "bootdef"
  crc   : "CRC16_Slow"
  fds   : "FullDuplexSerial" ' simple serial is missing some useful functions
  bsd   : "fsrw"

CON ' PinDefs
   _clkmode  = pd#_clkmode
   _xinfreq  = pd#_xinfreq

  BS = 8
  LF = $A
  CR = $D
  COGMAX = 8
  INTERPRETER_ADDRESS = $f004

DAT
  pad1        long 0
  bootconf    byte "bootconf.bin", 0
  
DAT
  openFail    byte CR,"Cant open ",0
  readFail    byte CR,"Cant read ",0
  memFail     byte CR,"No memory for ",0
  jazzed      byte "  jazzed  "
   
DAT ' Config Version 0.00 overlay .... the fields for boot beyond cfg_sdboot_fname except cfg_crc16 may change
{{
The config data section must always begin at $C0.
Add padding to make this happen.
}}
  pad2        long 0, 0, 0
  
  cfg_size                long $f0 ' keep as size-4 to exclude crc16      
  cfg_version             long $10001 ' Version number represented as $HHdd H is header version, d is data version
  cfg_sdpins              long 12 ' 
  cfg_baud_rate           long 115200
  cfg_sdboot_fname        long 0[4]  ' 16 chars
  cfg_pingrp1             byte "nnnnnnnn" ' offset $10: tv base, vga base, ntsc/pal, sdcard base, mouse dat, mouse clk, kb dat, kb clk
  cfg_pingrp2             byte "nnnnnnnn" ' ethernet pins: i_cs, i_sck, i_si, i_so, i_int
  cfg_pingrpRSV           long 0[12] ' reserved for more pin definitions  
  cfg_prodname            long 0[4]  ' 16 chars ... product name or copyright  
  cfg_userid              long 0[4]  ' 16 chars 
  cfg_passwd              long 0[4]  ' 16 chars 
  cfg_hostid              long 0[4]  ' 16 chars 
  cfg_ipv4_address        long 0 '   
  cfg_ipv4_netmask        long 0 ' 
  cfg_ipv4_gateway        long 0 ' 
  cfg_ipv4_dns            long 0 ' 
  cfg_macaddr             long 0, 0 ' 8 bytes ... only need 6 
  cfg_rsv2_data           long 0, 0 ' 2 longs
  cfg_bootp_fname         long 0[8] ' 32 chars     
  cfg_rsv3                long 0[3] 
  cfg_crc16               long $8b42 ' magic-number crc '
  

CON
  PROGLEN = 26300
  
VAR
  byte heap[PROGLEN]
  byte proghdr[16]
  long fdscog

PUB main | badcrc, rc, fname

  badcrc := getEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)

  if badcrc
    rc := fds.start(pd#SERIAL1_RX, pd#SERIAL1_TX, pd#SERIAL1_MODE, pd#SERIAL1_BAUD)
  else
    rc := fds.start(pd#SERIAL1_RX, pd#SERIAL1_TX, pd#SERIAL1_MODE, cfg_baud_rate)

  'badcrc := getEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)

  fdscog := rc
  waitcnt(clkfreq+cnt)

  print(string(CR,CR,"Bootloader started. " ))
  dec(@proghdr-@heap)
  print(string(" bytes free.",$d))
  waitcnt(clkfreq+cnt)

  if @cfg_size <> bd#BIOS_DBASE_ADDR
    print(string("Bad config address.",CR))
    repeat
  
  fname := @cfg_sdboot_fname
  'fname := string("www.bin")
  'badcrc := getEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)

  if badcrc <> 0
    getSdBasePin
    byte[fname] := 0
   
  rc := iskey
  if rc > 0
    puts(string("Key hit"))
    if (rc-1) == "C"                                                                 
      byte[fname] := 0
      print(string("Boot "))
      print(@bootconf)
    else
      dir
      print(string("Bootfile? "))
      waitcnt(clkfreq+cnt)
      fname := getString(@heap,20,CR)
  
  if strsize(fname) == 0 or byte[fname] == " "
    fname := @bootconf 
   
  'try to boot from sdcard
  rc := \boot(fname)
  ifnot rc
    repeat
  else
    print(string(CR,"Boot failed "))
    dec(rc)


PRI getSdBasePin | buf[10], ch, ii
  print(string("Bad config. Enter SD card basepin: "))
  byte[@cfg_sdpins+bd#SDCARD_BASEPIN_BYTE] := getPin(getString(@buf, 8, CR))
  'byte[@cfg_sdpins+bd#SDCARD_BASEPIN_BYTE] := conv.atoi(getString(@buf, 8, CR))

PRI getPin(s)
  result := 0
  repeat while (byte[s] => "0" and byte[s] =< "9")
    result := (10*result) + (byte[s++] - "0") '// calculate value

PRI getString(buf,lim,trm) | ch, ii
  ii := 0
  byte[buf] := 0
  repeat
    ch := getchar
    if ch == trm
      quit
    if ch == BS and ii > 0
      ii--
    else
      byte[buf+ii] := ch
      ii++
    byte[buf+ii] := 0
    putchar(ch)
    if ii => lim
      quit
  return buf

PRI dir | sdpin, sbuf[100]

  result := sdopen
  if result > -1
    bsd.opendir
    repeat while bsd.nextfile(@sbuf) == 0
      puts(@sbuf)

PRI boot(fname) | ii, sbuf[100], rc, len, sdpin, spinbin

  print(string(CR,"Boot from SD Card pin "))
  dec(byte[@cfg_sdpins+bd#SDCARD_BASEPIN_BYTE])

  if sdopen > -1
    spinbin := loadProgram(fname)

  if spinbin
    newline   ' last chance for output
    ' kill all cogs except main interpreter
    repeat ii from 0 to 7
      if ii <> cogid
        cogstop(ii)
    cognew(INTERPRETER_ADDRESS, spinbin+4)
    waitcnt(clkfreq/10+cnt)
    cogstop(cogid)
    repeat

PRI loadError(s,fname)
  print(s)
  puts(fname)
  bsd.pclose
  return 0
          
PRI loadProgram(fname) | rc,ii,plen,progptr

  rc := bsd.popen(fname, "r")
  if rc
    return loadError(@openFail,fname)
   
  repeat ii from 0 to 15
    byte[@proghdr][ii] := bsd.pgetc

  plen := word[@proghdr+$A] ' plen at stack base gives enough room for prog
   
  print(string(CR,"File "))
  print(fname)
  print(string(" size "))
  dec(plen)
  newline

  if plen > PROGLEN
    return loadError(@memFail,fname)
        
  bsd.pclose

  progptr := @heap
  progptr := (progptr + 4) & !3
  
  rc := bsd.popen(fname, "r")
  if rc
    return loadError(@openFail,fname)

  rc := bsd.pread(progptr, plen)
  
  if rc =< 0
    return loadError(@readFail,fname)

  bsd.pclose

  ' Fixup
  Repeat ii from 0 to 4
    Word[progptr+6+ii<<1] += progptr

  ' Clear varspace
  rc := word[@proghdr+$8]
  Repeat ii from rc to plen step 4
    long[progptr+ii] := 0
    
  long[progptr+plen-4] := $fff9ffff
  long[progptr+plen-8] := $fff9ffff

  return progptr


PRI getEEconfig(dptr,length) | mycrc, ii, rc
' read database ... already loaded with object DAT section. just verify CRC

  ' start with -1 to detect blank array
  mycrc := crc.get(@cfg_size, @cfg_crc16-@cfg_size, crc#CRC_INIT)

  if(long[dptr+length-1] <> mycrc) and fdscog
    print(string("Bad CRC "))
    hex(long[dptr+length-1],4)
    print(string(" Expected "))
    hex(mycrc,4)
    newline

  if(long[dptr+length-1] <> mycrc)
    return -1

  return 0    


PRI getChar | id
  return fds.rx

PRI iskey
  return fds.rxcheck

PRI newline
  fds.tx(CR)
  
PRI putchar(ch)
  fds.tx(ch)

PRI puts(str)
  fds.str(str)
  newline

PRI print(str)
  fds.str(str)

PRI dec(v)
  fds.dec(v)
    
PRI hex(v,len)
  fds.hex(v,len)

PRI sdopen | pin
  pin := byte[@cfg_sdpins+bd#SDCARD_BASEPIN_BYTE]
  result := \bsd.mount(pin)
  if result < 0
    print(string(CR,"Cant mount SD card "))
    dec(pin)
    newline

 